from .execution import Locker, Execution
from .msg import Verbosity
from glob2 import iglob
from pathlib import Path
import abc
import glob
import os
import pickle
import shutil


class CacheProvider(metaclass=abc.ABCMeta):

    def __init__(self, locker=None, verbosity=Verbosity.INFO, provider=None):
        self._locker = locker if locker else (provider._locker if provider else Locker(verbosity=verbosity))
        self._verbosity = verbosity
        self._provider = provider

    @abc.abstractmethod
    def prepare(self):
        pass

    @abc.abstractmethod
    def get_or_execute(self, execution, already_exclusive=False):
        pass

    @abc.abstractmethod
    def size(self):
        pass

    @abc.abstractmethod
    def is_valid_cache(self, execution):
        pass

    def lock(self, obj=None, already_exclusive=False):
        return EmptyContext() if already_exclusive else self._locker.lock(obj)

    @property
    def verbosity(self):
        return self._verbosity

    @property
    def provider(self):
        return self._provider

    @abc.abstractmethod
    def clear(self):
        pass


class InMemoryCacheProvider(CacheProvider):

    def __init__(self, verbosity=Verbosity.INFO, locker=None, max_entries=1000, provider=None):
        CacheProvider.__init__(self, locker, verbosity, provider)
        self._max_entries = max_entries
        self._cache = {}
        self._cache_priority = {}

    def prepare(self):
        if self._provider is not None:
            return self._provider.prepare()

    def get_or_execute(self, execution, already_exclusive=False):
        with self.lock(execution, already_exclusive):
            if self.is_valid_cache(execution):
                execution_name = execution.name
                result = self._cache[execution_name]
                self._cache_priority[execution_name] = self._cache_priority.get(execution_name, 0) + 1
                return False, result
            if self._provider is None:
                executed = True
                execution_result = execution()
            else:
                executed, execution_result = self._provider.get_or_execute(execution, already_exclusive=True)
            execution_name = execution.name
            if len(self._cache) >= self._max_entries:
                for cache_key, _ in sorted(self._cache_priority.items(), key=lambda x: x[1])[:self._max_entries // 2]:
                    del self._cache[cache_key]
                self._cache_priority = {}
            self._cache[execution_name] = execution_result
            self._cache_priority[execution_name] = self._cache_priority.get(execution_name, 0) + 1
            return executed, execution_result

    def size(self):
        return len(self._cache)

    def is_valid_cache(self, execution):
        return execution.name in self._cache and all([self.is_valid_cache(e) for e in execution.dependencies])

    def to_serializable(self):
        return {
            'max_entries': self._max_entries,
        }

    @staticmethod
    def from_serializable(serializable, verbosity):
        return InMemoryCacheProvider(verbosity, max_entries=serializable['max_entries'])

    def clear(self, recursively=True):
        self._cache = {}
        self._cache_priority = {}
        if recursively and self._provider is not None:
            self._provider.clear()


class StorageCacheProvider(CacheProvider):

    def __init__(self, storage, verbosity=Verbosity.INFO, locker=None, override=False, provider=None):
        CacheProvider.__init__(self, locker, verbosity, provider)
        self._storage = storage
        self._override = override
        self._time = None

    def prepare(self):
        with self.lock():
            self._storage.write_info(init=True)
            self._time = self._storage.read_info_time()
            self._storage.write_info(override_time=self._time)
            for _ in self._storage.read_executions():
                pass
            if self._provider is not None:
                self._provider.prepare()

    def get_or_execute(self, execution, already_exclusive=False):
        with self.lock(execution, already_exclusive):
            if self.is_valid_cache(execution):
                return False, self._storage.read_execution_result(execution)
            if self._provider is None:
                executed = True
                execution()
            else:
                executed, self._provider.get_or_execute(execution, already_exclusive=True)
            self._storage.write_execution_result(execution)
        return executed, self._storage.read_execution_result(execution)

    def size(self):
        with self.lock():
            return sum(1 for _ in self._storage.read_executions())

    def is_valid_cache(self, execution, reread=True):
        if reread:
            execution = self._storage.read_execution(execution)
            if execution is None:
                return False
        execution_time = self._storage.read_execution_time(execution)
        if self._override and execution_time < self._time:
            return False
        return execution_time >= self._get_execution_dependencies_max_time(execution) and all([self.is_valid_cache(d, reread=False) for d in execution.dependencies])

    def clear(self, recursively=True):
        self._storage.clear()
        if recursively and self._provider:
            self._provider.clear(recursively)

    def _get_execution_dependencies_max_time(self, execution):
        times = [self._storage.read_execution_time(d) for d in execution.dependencies]
        times = [(t if t is not None else float('inf')) for t in times]
        if len(times) == 0:
            return - float('inf')
        return max(times)


class Storage(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def delete_execution_result(self, execution):
        pass

    @abc.abstractmethod
    def write_execution_result(self, execution):
        pass

    @abc.abstractmethod
    def read_execution_result(self, execution):
        pass

    @abc.abstractmethod
    def read_execution_time(self, execution):
        pass

    @abc.abstractmethod
    def read_execution(self, execution):
        pass

    @abc.abstractmethod
    def read_info(self):
        pass

    @abc.abstractmethod
    def read_info_time(self):
        pass

    @abc.abstractmethod
    def write_info(self, **kwargs):
        pass

    @abc.abstractmethod
    def read_executions(self, function=None):
        pass

    @abc.abstractmethod
    def clear(self):
        pass


class FileStorage(Storage):

    def __init__(self, directory, verbosity=Verbosity.INFO):
        self._directory = directory
        self._verbosity = verbosity

    def delete_execution_result(self, execution):
        for filename in [self._get_filename(execution.name, 'execution.info.pickle'), self._get_filename(execution.name, 'execution.pickle')]:
            try:
                os.remove(filename)
            except OSError:
                pass

    def write_execution(self, execution):
        filename = self._get_filename(execution.name, 'execution.info.pickle', prepare=True)
        with open(filename, 'wb') as f:
            pickle.dump(execution.to_serializable(), f)

    def write_execution_ready(self, execution):
        filename = self._get_filename(execution.name, 'execution.ready', prepare=True)
        open(filename, 'a').close()

    def is_execution_ready(self, execution):
        return os.path.exists(self._get_filename(execution.name, 'execution.ready', prepare=False))

    def write_execution_result(self, execution):
        execution_result = execution()
        filename = self._get_filename(execution.name, 'execution.pickle', prepare=True)
        with open(filename, 'wb') as f:
            pickle.dump(execution_result, f)
        self.write_execution(execution)
        self.write_execution_ready(execution)

    def read_execution_result(self, execution):
        if not self.is_execution_ready(execution):
            return None
        filename = self._get_filename(execution.name, 'execution.pickle')
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def read_execution_time(self, execution):
        if not self.is_execution_ready(execution):
            return None
        filename = self._get_filename(execution.name, 'execution.pickle')
        return os.path.getmtime(filename)

    def read_execution(self, execution):
        if not self.is_execution_ready(execution):
            return None
        filename = self._get_filename(execution.name, 'execution.info.pickle')
        with open(filename, 'rb') as f:
            return Execution.from_serializable(pickle.load(f))

    def read_executions(self, function=None):
        if function is None:
            walker = Path(self._directory).glob(os.path.join('**', '*.execution.info.pickle'))
        else:
            walker = iglob(os.path.join(self._directory, function.name.replace('.', os.sep), '*.execution.info.pickle'))
        for path in walker:
            with open(str(path), 'rb') as f:
                yield Execution.from_serializable(pickle.load(f))

    def read_info(self):
        filename = self._get_filename('info', 'pickle')
        if not os.path.exists(filename):
            return {}
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def read_info_time(self):
        filename = self._get_filename('info', 'pickle')
        if not os.path.exists(filename):
            return None
        return os.path.getmtime(filename)

    def write_info(self, **kwargs):
        info = self.read_info()
        for key, value in kwargs.items():
            info[key] = value
        filename = self._get_filename('info', 'pickle', prepare=True)
        with open(filename, 'wb') as f:
            pickle.dump(info, f)

    def clear(self):
        for f in glob.iglob(os.path.join(self._directory, '*')):
            try:
                os.remove(f)
            except IsADirectoryError:
                shutil.rmtree(f, ignore_errors=True)

    def _get_filename(self, object_name, extension, prepare=False):
        filename = os.path.join(self._directory, '{}.{}'.format(object_name.replace('.', os.sep), extension))
        if prepare:
            directory = os.path.dirname(filename)
            if not os.path.exists(directory):
                os.makedirs(directory)
        return filename


class EmptyContext:

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



LAMBDA = lambda: 0


def is_lambda(fun):
    """
    Check whether the given function is a lambda function.

    .. testsetup::

        from spiderpig.func import is_lambda

    .. testcode::

        def not_lambda_fun(): ---------------------------------------------------------------------------------------------------------------------------------------------------------------- ---
        lambda_fun = lambda:

        print(
            is_lambda(not_lambda_fun),
            is_lambda(lambda_fun)
        )
    .. testoutput::

        False True

    Args:
        fun (function)

    Returns:
        bool: True if the given function is a lambda function, False otherwise
    """
    return isinstance(fun, type(LAMBDA)) and fun.__name__ == LAMBDA.__name__


def function_name(function):
    if function is None:
        return None
    if is_lambda(function):
        raise Exception('The function can not be a lambda function.')
    return '{}.{}'.format(function.__module__, function.__qualname__)



from .exceptions import ValidationError
from .msg import Verbosity
import argparse
import os


def get_argument_parser():
    p = argparse.ArgumentParser()

    p.add_argument(
        '--spiderpig-dir',
        action='store',
        dest='spiderpig_dir',
        default='.spiderpig'
    )
    p.add_argument(
        '--override-cache',
        action='store_true',
        dest='override_cache',
        default=False)
    p.add_argument(
        '--verbosity',
        action='store',
        dest='verbosity',
        type=int,
        default=Verbosity.INFO,
        choices=[Verbosity.INFO, Verbosity.DEBUG, Verbosity.VERBOSE, Verbosity.INTERNAL])
    p.add_argument(
        '--max-in-memory-entries',
        action='store',
        dest='max_in_memory_entries',
        default=1000
    )
    return p


def process_kwargs(parsed_kwargs):
    return {key: _convert_kwarg_value(val) for (key, val) in parsed_kwargs.items()}


def _convert_kwarg_value(val):
    if not isinstance(val, str):
        return val
    if val == 'False':
        return False
    elif val == 'True':
        return True
    else:
        try:
            return int(val)
        except ValueError:
            pass
        try:
            return float(val)
        except ValueError:
            return val
    return val


class Configuration:

    def __init__(self, configuration=None, **kwargs):
        self._kwargs = process_kwargs(kwargs)
        self._configuration = configuration

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise ValidationError('The key argument has to be a string.')
        if key in self._kwargs:
            return self._kwargs[key]
        if self._configuration is not None and key in self._configuration:
            return self._configuration[key]
        if key.upper() in os.environ:
            return _convert_kwarg_value(os.environ[key.upper()])
        raise KeyError(key)

    def __contains__(self, key):
        return key in self._kwargs or (self._configuration is not None and key in self._configuration) or key.upper() in os.environ

    def to_serializable(self):
        result = {
            'kwargs': dict(self._kwargs),
        }
        if self._configuration is not None:
            result['configuration'] = self._configuration.to_serializable()
        return result

    @staticmethod
    def from_serializable(serializable):
        configuration = None if 'configuration' not in serializable else Configuration.from_serializable(serializable['configuration'])
        return Configuration(configuration, **serializable['kwargs'])

    def __str__(self):
        return str(self._kwargs)



from .config import Configuration
from .exceptions import ValidationError, CyclicExecution
from .func import function_name
from .msg import Verbosity, print_debug
from clint.textui import indent
from collections import defaultdict
from functools import reduce
from glob import iglob
from threading import currentThread
from time import time
import filelock
import hashlib
import importlib
import inspect
import json
import os
import re
import tempfile


class Function:

    _dependencies = defaultdict(list)
    _dependency_names = defaultdict(set)

    def __init__(self, raw_function):
        self._raw_function = raw_function

    def add_dependency(self, function):
        name = self.name
        if function.name != name and function.name not in self._dependency_names[name]:
            self._dependencies[name].append(function)
            self._dependency_names[name].add(function.name)

    @property
    def arguments(self):
        if hasattr(self.raw_function, '__wrapped__'):
            return inspect.getargspec(self.raw_function.__wrapped__).args
        else:
            return inspect.getargspec(self.raw_function).args

    @property
    def defaults(self):
        if hasattr(self.raw_function, '__wrapped__'):
            return inspect.getargspec(self.raw_function.__wrapped__).defaults
        else:
            return inspect.getargspec(self.raw_function).defaults

    @property
    def dependencies(self):
        return list(Function._dependencies[self.name])

    @property
    def dependent_arguments(self):
        return set(self.arguments) | reduce(
            lambda a, b: a | b,
            [d.dependent_arguments for d in self.dependencies],
            set()
        )

    @staticmethod
    def from_name(function_name):
        matched = re.match('(.*)\.(\w+)', function_name)
        module = importlib.import_module(matched.groups()[0])
        return Function(getattr(module, matched.groups()[1]))

    @property
    def name(self):
        if not hasattr(self, '_name'):
            self._name = function_name(self.raw_function)
        return self._name

    @property
    def raw_function(self):
        return self._raw_function

    def to_serializable(self):
        name = self.name
        return {
            'function_name': name,
            'dependencies': [fun.to_serializable() for fun in Function._dependencies[name]],
        }

    @staticmethod
    def from_serializable(serializable):
        fun = Function.from_name(serializable['function_name'])
        for dep in serializable['dependencies']:
            fun.add_dependency(Function.from_serializable(dep))
        return fun

    @staticmethod
    def clear_dependencies():
        Function._dependencies = defaultdict(list)
        Function._dependency_names = defaultdict(set)

    def __call__(self, *args, **kwargs):
        return self.raw_function(*args, **kwargs)

    def __eq__(self, other):
        if not hasattr(other, 'raw_function'):
            return False
        else:
            return other.raw_function == self.raw_function

    def __str__(self):
        return self.name


class Execution:

    def __init__(self, function, configuration, verbosity=Verbosity.INFO, **kwargs):
        if isinstance(function, Function):
            self._function = function
        elif isinstance(function, str):
            self._function = Function.from_name(function)
        else:
            self._function = Function(function)
        self._configuration = configuration
        self._kwargs = kwargs
        self._value = None
        self._executed = False
        self._dependencies = []
        self._time = None
        self._verbosity = verbosity

    def add_dependency(self, execution):
        if execution.name not in {e.name for e in self._dependencies}:
            self._dependencies.append(execution)

    @property
    def context_kwargs(self):
        context_kwargs = {}
        for arg in self.function.dependent_arguments:
            if arg in self._configuration and arg not in self.kwargs:
                context_kwargs[arg] = self._configuration[arg]
        return context_kwargs

    @property
    def dependencies(self):
        return list(self._dependencies)

    @property
    def function(self):
        return self._function

    @property
    def kwargs(self):
        return dict(self._kwargs)

    @property
    def name(self):
        return '{}.{}'.format(self.function.name, hashlib.sha1((
            self.function.name + _serialize(self.kwargs) + _serialize(self.context_kwargs)
        ).encode()).hexdigest())

    def to_serializable(self):
        return {
            'function': self.function.to_serializable(),
            'configuration': self._configuration.to_serializable(),
            'kwargs': self.kwargs,
            'dependencies': [e.to_serializable() for e in self._dependencies],
        }

    @staticmethod
    def from_serializable(serializable, verbosity=Verbosity.INFO):
        execution = Execution(
            function=Function.from_serializable(serializable['function']),
            configuration=Configuration.from_serializable(serializable['configuration']),
            verbosity=verbosity,
            **serializable['kwargs']
        )
        for d in serializable['dependencies']:
            execution.add_dependency(Execution.from_serializable(d, verbosity=verbosity))
        return execution

    @property
    def time(self):
        return self._time

    def __call__(self):
        if not self._executed:
            time_before = time()
            kwargs = self.kwargs
            if 'verbosity' in self._function.arguments:
                kwargs['verbosity'] = self._verbosity
            self._value = self.function(**kwargs)
            self._time = time() - time_before
            self._executed = True
            if self._verbosity > Verbosity.DEBUG:
                print_debug('execution {0} took {1:.3f} seconds'.format(self.name, self._time))
                with indent(4):
                    for key, val in sorted(self.kwargs.items()):
                        print_debug('{}: {}'.format(key, val))
        return self._value

    def __eq__(self, other):
        return self.function == other.function and self.kwargs == other.kwargs and self.context_kwargs == other.context_kwargs

    def __str__(self):
        if not hasattr(self, '_str'):
            defaults = self.function.defaults
            if not isinstance(defaults, tuple):
                defaults = [defaults]

            kwargs = dict(zip(reversed(self.function.arguments), reversed(defaults)))
            kwargs.update(self.kwargs)
            self._str = '{}({})'.format(
                self.function.name,
                ','.join(
                    '{}={}'.format(key, value.fingerprint() if hasattr(value, 'fingerprint') else value)
                    for key, value in sorted(kwargs.items())
                )
            )
        return self._str


class ExecutionContext:

    def __init__(self, configuration=None, cache_provider=None, verbosity=Verbosity.INFO, locker=None):
        self._cache_provider = cache_provider
        self._execution_chain = defaultdict(list)
        self._configuration = configuration if configuration else Configuration()
        self._execution_count = defaultdict(lambda: 0)
        self._verbosity = verbosity
        self._locker = locker if locker else (cache_provider._locker if cache_provider else Locker())

    @property
    def configuration(self):
        return self._configuration

    def execute(self, raw_function, *args, use_cache=True, **kwargs):
        function = Function(raw_function)
        arg_names = function.arguments
        args_kwargs = dict(zip(arg_names, args))
        kwarg_intersection = set(args_kwargs.keys()) & set(kwargs)
        if len(kwarg_intersection) > 0:
            raise ValidationError('Can not pass value for {} as both argument and key-word argument.'.format(kwarg_intersection))
        kwargs.update(args_kwargs)
        exec_kwargs = self._get_kwargs(function, **kwargs)
        execution = Execution(function, self._configuration, verbosity=self.verbosity, **exec_kwargs)
        execution_chain = self._execution_chain[currentThread()]
        if execution in execution_chain:
            raise CyclicExecution('There is an execution cycle: {} -> {}'.format(
                execution.function.name,
                [str(e) for e in execution_chain]
            ))
        for execution_segment in execution_chain:
            execution_segment.add_dependency(execution)
            execution_segment.function.add_dependency(execution.function)
        execution_chain.append(execution)
        try:
            executed, result = (True, execution()) if (self._cache_provider is None or not use_cache) else self._cache_provider.get_or_execute(execution)
            self._execution_count[str(execution)] += executed
        finally:
            execution_chain.pop()
        return result

    def _get_kwargs(self, function, **cache_kwargs):
        valid_args = function.arguments
        for execution_segment in self._execution_chain[currentThread()][::-1]:
            for key, value in execution_segment.kwargs.items():
                if key in valid_args and key not in cache_kwargs:
                    cache_kwargs[key] = value
        for arg in valid_args:
            if arg in self._configuration and arg not in cache_kwargs:
                cache_kwargs[arg] = self._configuration[arg]
        return cache_kwargs

    @property
    def verbosity(self):
        return self._verbosity

    def count_executions(self, function, **kwargs):
        return self._execution_count[str(Execution(function, {}, verbosity=self.verbosity, **kwargs))]


def _serialize(x):
    if hasattr(x, 'fingerprint'):
        return json.dumps([
            '{}.{}'.format(x.__class__.__module__, x.__class__.__name__),
            x.fingerprint()
        ], sort_keys=True)
    try:
        return json.dumps(x, sort_keys=True)
    except TypeError:
        if isinstance(x, dict):
            return json.dumps(sorted([(k, _serialize(v)) for (k, v) in x.items()]), sort_keys=True)
        elif isinstance(x, list):
            return json.dumps([_serialize(v) for v in x])
        else:
            return json.dumps({
                'class': str(x.__class__),
                'data': x.__dict__,
            }, sort_keys=True)


class Locker:

    def __init__(self, directory=None, verbosity=Verbosity.INFO):
        self._directory = tempfile.mkdtemp() if directory is None else directory
        self._verbosity = verbosity
        if not os.path.exists(self._directory):
            os.makedirs(self._directory)

    def clear(self):
        for filename in iglob(os.path.join(self._directory, '*.lock')):
            try:
                os.remove(filename)
            except OSError:
                pass

    def lock(self, obj=None):
        name = 'spiderpig.global' if obj is None else obj.name
        return LockWrapper(obj, filelock.FileLock(os.path.join(self._directory, '{}.lock'.format(name))), self._verbosity)

    def to_serializable(self):
        return {'directory': self._directory}

    @staticmethod
    def from_serializable(serializable):
        return Locker(serializable['directory'])


class LockWrapper:

    def __init__(self, obj, lock, verbosity):
        self._lock = lock
        self._obj = obj
        self._verbosity = verbosity

    def __enter__(self):
        return self._lock.__class__.__enter__(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._lock.__class__.__exit__(self, exc_type, exc_val, exc_tb)

    def __getattr__(self, attr):
            self._lock.__class__
            orig_attr = self._lock.__getattribute__(attr)
            if self._verbosity < Verbosity.INTERNAL or attr not in ['acquire', 'release']:
                return orig_attr
            current_frame = inspect.currentframe()
            call_frame = inspect.getouterframes(current_frame, 3)
            line = call_frame[3]

            def hooked(*args, **kwargs):
                if attr == 'acquire':
                    print_debug('{}:{} trying to lock {}'.format(line[1], line[2], self._obj))
                result = orig_attr(*args, **kwargs)
                if attr == 'acquire':
                    print_debug('{}:{} locking {}'.format(line[1], line[2], self._obj))
                else:
                    print_debug('{}:{} unlocking {}'.format(line[1], line[2], self._obj))
                return self if result is self._obj else result
            return hooked



from pytest import raises
from spiderpig.config import Configuration
from spiderpig.exceptions import ValidationError
from spiderpig.execution import Function, ExecutionContext, Execution
from spiderpig.func import function_name


def test_function():
    fun = Function.from_name('spiderpig.func.function_name')
    assert fun is not None
    assert fun.arguments == ['function']
    assert fun(function_name) == 'spiderpig.func.function_name'
    assert fun(function=Function.from_name) == 'spiderpig.execution.Function.from_name'
    dep_fun = Function.from_name('spiderpig.func.is_lambda')
    fun.add_dependency(dep_fun)
    assert fun.dependencies == [dep_fun]
    assert Function.from_name('spiderpig.func.function_name').dependencies == [dep_fun]
    assert fun == Function.from_name('spiderpig.func.function_name')
    assert fun.dependent_arguments == {'function', 'fun'}
    assert fun.name == 'spiderpig.func.function_name'
    assert fun.raw_function is function_name
    s_fun = Function.from_serializable(fun.to_serializable())
    assert s_fun == fun
    with raises(ImportError):
        Function.from_name('aaa.bbb')


def test_execution():
    reset_calls()
    execution = Execution(fun_a, Configuration(), a=1)
    s_execution = Execution.from_serializable(execution.to_serializable())
    assert s_execution == execution
    assert execution() == 1
    assert get_calls('a') == [{'a': 1}]


def test_execution_context():
    reset_calls()
    context = ExecutionContext(Configuration(a=2))
    assert context.execute(fun_a) == 2
    assert get_calls('a') == [{'a': 2}]
    assert context.count_executions(fun_a, a=2) == 1
    assert context.count_executions(fun_a, a=1) == 0
    reset_calls()
    assert context.execute(fun_a, a=3) == 3
    assert get_calls('a') == [{'a': 3}]
    with raises(ValidationError):
        context.execute(fun_a, 1, a=2)


_CALLS = {}


def reset_calls():
    global _CALLS
    _CALLS = {}


def get_calls(name):
    global _CALLS
    return _CALLS.get(name, [])


def fun_a(a):
    global _CALLS
    calls = _CALLS.get('a', [])
    calls.append({'a': a})
    _CALLS['a'] = calls
    return a



from multiprocessing.pool import ThreadPool
from pytest import raises
from spiderpig.msg import Verbosity
from time import sleep
import os
import spiderpig
import tempfile


def test_configured():
    with spiderpig.spiderpig(a=1, verbosity=Verbosity.INTERNAL):
        assert fun() == (1, 3)
        with spiderpig.configuration(a=2, b=3):
            assert fun() == (2, 5)
        assert fun() == (1, 3)
        assert fun(a=2) == (2, 4)
    with spiderpig.spiderpig(verbosity=Verbosity.INTERNAL):
        try:
            os.environ['A'] = '10'
            assert fun() == (10, 12)
            os.environ['B'] = '3'
            assert fun() == (10, 13)
            assert fun(a=1) == (1, 4)
            with spiderpig.configuration(a=2):
                assert fun() == (2, 5)
        finally:
            if 'A' in os.environ:
                del os.environ['A']
            if 'B' in os.environ:
                del os.environ['B']


def test_cached():
    cache_dir = tempfile.mkdtemp()
    with spiderpig.spiderpig(cache_dir, a=1, verbosity=Verbosity.INTERNAL):
        assert cached_fun() == (1, 3, 10)
        assert spiderpig.execution_context().count_executions(cached_fun) == 1
        assert spiderpig.execution_context().count_executions(cached_fun_a, a=1) == 1
        assert spiderpig.execution_context().count_executions(cached_fun_b) == 1
        assert spiderpig.execution_context().count_executions(cached_fun_c) == 1
    with spiderpig.spiderpig(cache_dir, a=1, verbosity=Verbosity.INTERNAL):
        assert cached_fun() == (1, 3, 10)
        assert spiderpig.execution_context().count_executions(cached_fun) == 0
        assert spiderpig.execution_context().count_executions(cached_fun_a, a=1) == 0
        assert spiderpig.execution_context().count_executions(cached_fun_b) == 0
        assert spiderpig.execution_context().count_executions(cached_fun_c) == 0
    with spiderpig.spiderpig(cache_dir, override_cache=True, a=1, verbosity=Verbosity.INTERNAL):
        assert cached_fun_a() == 1
        assert cached_fun_a() == 1
        assert spiderpig.execution_context().count_executions(cached_fun_a, a=1) == 1
    with spiderpig.spiderpig(cache_dir, a=1, verbosity=Verbosity.INTERNAL):
        assert cached_fun() == (1, 3, 10)
        assert spiderpig.execution_context().count_executions(cached_fun) == 1
        assert spiderpig.execution_context().count_executions(cached_fun_a, a=1) == 0
        assert spiderpig.execution_context().count_executions(cached_fun_b) == 1
        assert spiderpig.execution_context().count_executions(cached_fun_c) == 0
    with spiderpig.spiderpig(cache_dir, a=2, verbosity=Verbosity.INTERNAL):
        assert cached_fun() == (2, 4, 10)
        assert spiderpig.execution_context().count_executions(cached_fun) == 1
        assert spiderpig.execution_context().count_executions(cached_fun_a, a=2) == 1


def test_in_memory_entries():
    cache_dir = tempfile.mkdtemp()
    with spiderpig.spiderpig(cache_dir, max_in_memory_entries=10):
        for i in range(100):
            assert cached_fun_a(a=i) == i
        assert spiderpig.cache_provider().size() <= 10
    with spiderpig.spiderpig(cache_dir, max_in_memory_entries=10):
        for i in range(100):
            assert cached_fun_a(a=i) == i
            assert spiderpig.execution_context().count_executions(cached_fun_a, a=i) == 0
            assert spiderpig.cache_provider().size() <= 10
    cache_dir = tempfile.mkdtemp()
    spiderpig.init(cache_dir, max_in_memory_entries=10)
    for i in range(100):
        assert cached_fun_a(a=i) == i
    assert spiderpig.cache_provider().size() <= 10
    spiderpig.terminate()
    spiderpig.init(cache_dir, max_in_memory_entries=10)
    for i in range(100):
        assert cached_fun_a(a=i) == i
        assert spiderpig.execution_context().count_executions(cached_fun_a, a=i) == 0
        assert spiderpig.cache_provider().size() <= 10


def test_exceptions():
    with spiderpig.spiderpig(verbosity=Verbosity.INTERNAL):
        with raises(RandomError):
            errored()
        with raises(RandomError):
            errored()


def test_concurrency():
    cache_dir = tempfile.mkdtemp()
    pool = ThreadPool(4)
    with spiderpig.spiderpig(cache_dir, verbosity=Verbosity.INTERNAL):
        assert pool.map(waiting_fun, [None] * 4) == [True] * 4
        assert spiderpig.execution_context().count_executions(waiting_fun) == 1


@spiderpig.configured()
def fun_a(a=None):
    return a


@spiderpig.configured()
def fun_b(b=2):
    return b + fun_a()


@spiderpig.configured()
def fun(a=None):
    return fun_a(), fun_b()


@spiderpig.cached()
def cached_fun_a(a=None):
    return a


@spiderpig.cached()
def cached_fun_b(b=2):
    return b + cached_fun_a()


@spiderpig.cached()
def cached_fun_c(c=10):
    return c


@spiderpig.cached()
def cached_fun():
    return cached_fun_a(), cached_fun_b(), cached_fun_c()


class RandomError(Exception):
    pass


@spiderpig.cached()
def errored():
    raise RandomError()


@spiderpig.cached()
def waiting_fun():
    sleep(1)
    return True



from spiderpig.cache import InMemoryCacheProvider, FileStorage, StorageCacheProvider
from spiderpig.execution import ExecutionContext, Locker
from spiderpig.msg import Verbosity
from spiderpig.tests.test_execution import reset_calls, get_calls, fun_a
import tempfile


def test_in_memory_cache():
    reset_calls()
    provider = InMemoryCacheProvider(max_entries=10)
    context = ExecutionContext(cache_provider=provider)
    for i in range(2):
        for _ in range(2):
            assert context.execute(fun_a, i) == i
    assert get_calls('a') == [{'a': i} for i in range(2)]
    assert provider.size() == 2
    provider.clear()
    for i in range(2):
        for _ in range(2):
            assert context.execute(fun_a, i) == i
    assert get_calls('a') == [{'a': i} for _ in range(2) for i in range(2)]
    provider.clear()
    for i in range(20):
        assert context.execute(fun_a, i) == i
        assert provider.size() <= 10


def test_storage_cache():
    reset_calls()
    storage = FileStorage(tempfile.mkdtemp())
    provider = StorageCacheProvider(storage=storage, locker=Locker(verbosity=Verbosity.INTERNAL))
    context = ExecutionContext(cache_provider=provider)
    for i in range(2):
        for _ in range(2):
            assert context.execute(fun_a, i) == i
    assert get_calls('a') == [{'a': i} for i in range(2)]
    assert provider.size() == 2
    provider.clear()
    for i in range(2):
        for _ in range(2):
            assert context.execute(fun_a, i) == i
    assert get_calls('a') == [{'a': i} for _ in range(2) for i in range(2)]


def test_storage_integration():
    reset_calls()
    storage = FileStorage(tempfile.mkdtemp())
    storage_provider = StorageCacheProvider(storage=storage, locker=Locker(verbosity=Verbosity.INTERNAL))
    in_memory_provider = InMemoryCacheProvider(provider=storage_provider)
    context = ExecutionContext(cache_provider=in_memory_provider)
    for i in range(2):
        for _ in range(2):
            assert context.execute(fun_a, i) == i
    assert get_calls('a') == [{'a': i} for i in range(2)]
    assert storage_provider.size() == 2
    assert in_memory_provider.size() == 2
    in_memory_provider.clear(recursively=False)
    assert storage_provider.size() == 2
    for i in range(2):
        for _ in range(2):
            assert context.execute(fun_a, i) == i
    assert get_calls('a') == [{'a': i} for i in range(2)]
    assert storage_provider.size() == 2
    assert in_memory_provider.size() == 2
    storage_provider.clear()
    for i in range(2):
        for _ in range(2):
            assert context.execute(fun_a, i) == i
    assert get_calls('a') == [{'a': i} for i in range(2)]


def test_storage_with_recursion():
    reset_calls()
    storage = FileStorage(tempfile.mkdtemp())
    storage_provider = StorageCacheProvider(storage=storage, locker=Locker(verbosity=Verbosity.INTERNAL))
    in_memory_provider = InMemoryCacheProvider(provider=storage_provider)
    context = ExecutionContext(cache_provider=in_memory_provider)
    assert context.execute(factorial, 6) == 720


def factorial(n):
    if n <= 1:
        return 1
    return n * factorial(n - 1)



from clint.textui import colored, puts


class Verbosity:
    INFO = 0
    DEBUG = 1
    VERBOSE = 2
    INTERNAL = 3


def print_debug(msg):
    puts(colored.yellow(str(msg)))


def print_info(msg):
    puts(colored.blue(str(msg)))


def print_success(msg):
    puts(colored.green(str(msg)))


def print_error(msg):
    puts(colored.red(str(msg)))



import inspect
import pkgutil
from spiderpig import msg


def register_submodule_command(subparsers, submodule, namespace=None):
    if 'command_name' in dir(submodule):
        command_name = submodule.command_name()
    else:
        command_name = submodule.__name__.split('.')[-1].replace('_', '-')
    command_name = command_name if namespace is None else '{}-{}'.format(namespace, command_name)
    subparser = subparsers.add_parser(command_name, help=submodule.__doc__)
    if 'init_parser' in dir(submodule):
        submodule.init_parser(subparser)
    else:
        args, varargs, keywords, defaults = inspect.getargspec(submodule.execute)

        def transform(x):
            return [] if x is None else x
        kwargs = dict(zip(args[-len(transform(defaults)):], transform(defaults)))
        for argname, default in kwargs.items():
            if isinstance(default, bool):
                used_argname = argname if not default else 'not-{}'.format(argname)
                action = 'store_true' if not default else 'store_false'
                subparser.add_argument('--{}'.format(used_argname.replace('_', '-')), action=action, required=False, help='default: {}'.format(default), dest=argname)
            else:
                subparser.add_argument('--{}'.format(argname.replace('_', '-')), action='store', default=default, required=False, help='default: {}'.format(default), dest=argname)
        for argname in args:
            if argname not in kwargs:
                subparser.add_argument('--{}'.format(argname), action='store', required=True)
    subparser.set_defaults(func=submodule.execute)


def register_submodule_commands(subparsers, package, namespace=None):
    prefix = package.__name__ + "."
    for importer, module_name, ispkg in pkgutil.iter_modules(package.__path__, prefix):
        if not ispkg:
            submodule = importer.find_module(module_name).load_module(module_name)
            if not is_submodule_command(submodule):
                continue
            register_submodule_command(subparsers, submodule, namespace=namespace)


def is_submodule_command(submodule):
    return hasattr(submodule, 'execute')


def execute(args):
    if 'func' not in args:
        msg.print_error('You have to choose subcommand!')
        return
    func = args['func']
    allowed_args = inspect.getargspec(func).args
    func_args = {key: value for (key, value) in args.items() if key in allowed_args}
    func(**func_args)



from spiderpig.execution import Function
from clint.textui import indent
import spiderpig.msg as msg
import spiderpig


def execute(function_name=None):
    storage = spiderpig.storage()
    function = Function.from_name(function_name) if function_name else None
    msg.print_info('Available cached executions:')
    with indent(4):
        for execution in sorted(storage.read_executions(function), key=lambda e: e.__str__()):
            msg.print_info('{}: {}'.format(execution, execution.name))



from . import cache
from . import commands, config
from . import execution
from .exceptions import ValidationError, NotInitialized
from .msg import Verbosity
from contextlib import ContextDecorator
from functools import wraps
import argcomplete
import json
import tempfile
import yaml


__VERSION__ = '2.3.0'


_EXECUTION_CONTEXT = None
_CACHE_PROVIDER = None
_STORAGE = None


class spiderpig(ContextDecorator):

    """
    Decorator/context initializing spiderpig.

    Examples
    --------

        >>> @configured()
        ... def fun_a(a=None):
        ...    print('A:', a)
        ...
        >>> @configured()
        ... def fun_b(a=None, b=None):
        ...    fun_a()
        ...    print('B:', b)
        ...
        >>> cache_dir = tempfile.mkdtemp()
        >>> with spiderpig(cache_dir, a=1, b=2):
        ...    fun_b()
        ...    fun_b(a=10, b=20)
        ...
        A: 1
        B: 2
        A: 10
        B: 20

    See also
    --------
    init
    """

    def __init__(self, directory=None, override_cache=False, verbosity=Verbosity.INFO, max_in_memory_entries=1000, config_file=None, **global_kwargs):
        """
        Initialize spiderpig for using it out of command-line tool.

        Parameters
        ----------
        directory: str
            path to a directory used for spiderpig auxiliary files and cache
        override_cache: bool, default False
            True if you want to recompute all spiderpig functions
            regardless of whether there is valid cache available, otherwise False
        verbosity: int, default 0
            increase verbosity level
        max_in_memory_entries: int, default 1000
            maximal number of entries in in-memory cache
        config_file: str
            path to the YAML/JSON file containing key-word parameters to
            override the global configuration
        global_kwargs: dict
            key-word arguments passed to spiderpig functions
        """
        self._directory = directory
        self._override_cache = override_cache
        self._verbosity = verbosity
        self._max_in_memory_entries = max_in_memory_entries
        self._global_kwargs = global_kwargs
        self._config_file = config_file

    def __enter__(self):
        init(self._directory, self._override_cache, self._verbosity, self._max_in_memory_entries, self._config_file, **self._global_kwargs)

    def __exit__(self, *exc):
        terminate()


def init(directory=None, override_cache=False, verbosity=Verbosity.INFO, max_in_memory_entries=1000, config_file=None, **global_kwargs):
    """
    Initialize spiderpig for using it out of command-line tool.

    Parameters
    ----------
    directory: str
        path to a directory used for spiderpig auxiliary files and cache
    override_cache: bool, default False
        True if you want to recompute all spiderpig functions
        regardless of whether there is valid cache available, otherwise False
    verbosity: int, default 0
        increase verbosity level
    max_in_memory_entries: int, default 1000
        maximal number of entries in in-memory cache
    config_file: str
        path to the YAML/JSON file containing key-word parameters to
        override the global configuration
    global_kwargs: dict
        key-word arguments passed to spiderpig functions

    Examples
    --------

        >>> @configured()
        ... def fun_a(a=None):
        ...    print('A:', a)
        ...
        >>> @configured()
        ... def fun_b(a=None, b=None):
        ...    fun_a()
        ...    print('B:', b)
        ...
        >>> cache_dir = tempfile.mkdtemp()
        >>> init(cache_dir, a=1, b=2)
        >>> fun_b()
        A: 1
        B: 2
        >>> fun_b(a=10, b=20)
        A: 10
        B: 20
    """
    global _EXECUTION_CONTEXT
    global _CACHE_PROVIDER
    global _STORAGE
    if config_file is not None:
        with open(config_file, 'r') as f:
            from_config_file = json.load(f.read()) if config_file.endswith('.json') else yaml.load(f.read())
            for key, value in from_config_file.items():
                if hasattr(value, '__len__') and not isinstance(value, str):
                    raise ValidationError('Config "{} ({})" is not scalar.'.format(key, value))
            from_config_file.update(global_kwargs)
            global_kwargs= from_config_file
    if directory is None:
        _CACHE_PROVIDER = cache.InMemoryCacheProvider(max_entries=max_in_memory_entries)
    else:
        _STORAGE = cache.FileStorage(directory if directory else tempfile.mkdtemp())
        _CACHE_PROVIDER = cache.InMemoryCacheProvider(
            provider=cache.StorageCacheProvider(
                storage=_STORAGE, verbosity=verbosity, override=override_cache
            ),
            max_entries=max_in_memory_entries
        )
    _CACHE_PROVIDER.prepare()
    _EXECUTION_CONTEXT = execution.ExecutionContext(
        configuration=config.Configuration(**global_kwargs),
        cache_provider=_CACHE_PROVIDER,
        verbosity=verbosity
    )


def terminate():
    """
    Terminates spiderpig and clears execution contexts and other auxiliary
    services.
    """
    global _EXECUTION_CONTEXT
    global _CACHE_PROVIDER
    global _STORAGE

    _EXECUTION_CONTEXT = None
    _CACHE_PROVIDER = None
    _STORAGE
    execution.Function.clear_dependencies()


class configuration:

    """
    Context overriding global configuration (key-word arguments) passed to all
    spiderpig functions. The configuration can be given directly as key-word
    arguments or loaded from configuration YAML/JSON file.

    Examples
    --------

        >>> @configured()
        ... def fun(a=None):
        ...     print('A:', a)
        ...
        >>> cache_dir = tempfile.mkdtemp()
        >>> with spiderpig(cache_dir, a=1):
        ...     fun()
        ...     with configuration(a=2):
        ...         fun()
        ...     fun()
        ...
        A: 1
        A: 2
        A: 1
    """

    def __init__(self, config_file=None, **config):
        """
        Override current global configuration.

        Parameters
        ----------
        config_file: str
            path to the YAML/JSON file containing key-word parameters to override
        config: dict
            key-word parameters to override
        """
        self._config = config
        if config_file is not None:
            with open(config_file, 'r') as f:
                from_config_file = json.load(f.read()) if config_file.endswith('.json') else yaml.load(f.read())
                for key, value in from_config_file.items():
                    if hasattr(value, '__len__') and not isinstance(value, str):
                        raise ValidationError('Config "{} ({})" is not scalar.'.format(key, value))
                from_config_file.update(self._config)
                self._config = from_config_file

    def __enter__(self):
        if len(self._config) == 0:
            return
        self._current_exec_context = execution_context()
        global _EXECUTION_CONTEXT
        _EXECUTION_CONTEXT = execution.ExecutionContext(
            configuration=config.Configuration(
                configuration=None if self._current_exec_context is None else self._current_exec_context.configuration,
                **self._config
            ),
            cache_provider=_CACHE_PROVIDER,
            verbosity=self._current_exec_context.verbosity
        )
        return self

    def __exit__(self, *exc):
        if len(self._config) == 0:
            return
        global _EXECUTION_CONTEXT
        _EXECUTION_CONTEXT = self._current_exec_context


def storage():
    """
    Retrieve the current storage used by spiderpig to persist cache. If
    spiderpig is initialized without directory, there is no storage and this
    function returns None.

    Returns
    -------
    storage currently used by spiderpig to persist cache
    """
    global _STORAGE
    if _STORAGE is None:
        raise NotInitialized('The storage is not initialized.')
    return _STORAGE


def cache_provider():
    """
    Retrieve the current cache provider.
    """
    global _CACHE_PROVIDER
    if _CACHE_PROVIDER is None:
        raise NotInitialized('The cache provider is not initialized.')
    return _CACHE_PROVIDER


def execution_context():
    """
    Retrieve the current execution context.
    """
    global _EXECUTION_CONTEXT
    if _EXECUTION_CONTEXT is None:
        raise NotInitialized('The execution context is not initialized.')
    return _EXECUTION_CONTEXT


class configured:

    """
    Decorator used to annotate spiderpig functions. Parameters for these
    functions are injected from the global configuration.

    See also
    --------
    spiderpig
    init
    cached
    """

    def __init__(self, cached=False, **config):
        """
        Create a decorator instance.

        Parameters
        ----------
        cached: bool, default False
            turn on caching
        config: dict
            key-word parameters to override the global configuration
        """
        self._cached = cached
        self._config = config

    def __call__(self, func):

        @wraps(func)
        def _wrapper(*args, **kwargs):
            def_args = execution.Function(func).arguments
            kwargs.update(dict(zip(def_args, args)))
            with configuration(**self._config):
                return execution_context().execute(func, use_cache=self._cached, **kwargs)

        return _wrapper


class cached(configured):

    """
    Decorator used to annotate cached spiderpig functions. Parameters for these
    functions are injected from the global configuration. Executions are cached
    based on the values of parameters of the given function and its
    dependencies.

    Examples
    --------

        >>> @cached()
        ... def fun(a=None):
        ...     print('executed')
        ...     return a
        ...
        >>> cache_dir = tempfile.mkdtemp()
        >>> with spiderpig(cache_dir, a=1):
        ...     print(fun())
        ...     print(fun())
        ...     print(fun(2))
        ...
        executed
        1
        1
        executed
        2

    See also
    --------
    spiderpig
    init
    configured
    """

    def __init__(self, **config):
        """
        Create a decorator instance.

        Parameters
        ----------
        cached: bool, default False
            turn on caching
        config: dict
            key-word parameters to override the global configuration
        """
        super().__init__(cached=True, **config)


def run_cli(command_packages=None, namespaced_command_packages=None, argument_parser=None, setup_functions=None):
    """
    Run spiderpig command-line application.

    Args:
        command_packages: list of packages containing modules representing
            command-line commands
        namespaced_command_packages: dictionary with packages containing
            modules representing command-line commands; keys are used as
            prefixes for the commands
        argument_parser: custom argument parser (argparse.ArgumentParser instance)
        setup_functions: functions invoked before the main command is executed
    """
    import spiderpig.commands.common as spcommon
    if command_packages is None:
        command_packages = []
    if namespaced_command_packages is None:
        namespaced_command_packages = {}
    if setup_functions is None:
        setup_functions = []
    parser = config.get_argument_parser() if argument_parser is None else argument_parser
    subparsers = parser.add_subparsers()
    for command_package in command_packages:
        commands.register_submodule_commands(subparsers, command_package)
    for command_namespace, command_package in namespaced_command_packages.items():
        commands.register_submodule_commands(subparsers, command_package, namespace=command_namespace)
    commands.register_submodule_commands(subparsers, spcommon, namespace='spiderpig')
    argcomplete.autocomplete(parser)
    args = vars(parser.parse_args())

    args = config.process_kwargs(args)
    with spiderpig(args['spiderpig_dir'], **{k: v for (k, v) in args.items() if k != 'func'}):
        for setup_fun in setup_functions:
            execution_context().execute(setup_fun, False)
        commands.execute(args)



from setuptools import setup


setup(
    name='spiderpig',
    version='2.3.0',
    description='Caching and execution library for data analysis.',
    author='Jan Papousek',
    author_email='jan.papousek@gmail.com',
    include_package_data=True,
    packages=[
        'spiderpig',
        'spiderpig.commands',
        'spiderpig.commands.common',
        'spiderpig.tests',
    ],
    install_requires=[
        'argcomplete',
        'clint',
        'filelock',
        'glob2',
        'html2text',
        'PyYAML',
    ],
    license='MIT',
)
